unit Main;

interface

uses
  System.SysUtils, System.Types, System.UITypes, System.Classes, System.Variants,
  FMX.Types, FMX.Controls, FMX.Forms, FMX.Dialogs, FMX.ListBox, FMX.Menus,
  FMX.Edit, Common;

type
  TMainForm = class(TForm)
    MainMenu: TMainMenu;
    FileMenu: TMenuItem;
    HelpMenu: TMenuItem;
    ExitMenu: TMenuItem;
    AboutMenu: TMenuItem;
    TypeLabel: TLabel;
    TypeCombo: TComboBox;
    MetricUnitLabel: TLabel;
    MetricCombo: TComboBox;
    MetricAmountLabel: TLabel;
    MetricEdit: TEdit;
    UsUnitLabel: TLabel;
    UsCombo: TComboBox;
    UsAmountLabel: TLabel;
    UsEdit: TEdit;
    StatusBar: TStatusBar;
    procedure FormCreate(Sender: TObject);
    procedure TypeComboChange(Sender: TObject);
    procedure MetricEditChange(Sender: TObject);
    procedure UsEditChange(Sender: TObject);
    procedure ExitMenuClick(Sender: TObject);
    procedure AboutMenuClick(Sender: TObject);

  private
    FUpdating: Boolean;

    function GetQuantity: TQuantity;
    procedure SetQuantity(value: TQuantity);

    function GetMetricLength: TLaMetricLength;
    function GetUsLength: TLaUsLength;

    function GetMetricTemperature: TLaMetricTemperature;
    function GetUsTemperature: TLaUsTemperature;

    function GetMetricValue: Double;
    procedure SetMetricValue(value: Double);

    function GetUsValue: Double;
    procedure SetUsValue(value: Double);

  public
    property Quantity: TQuantity read GetQuantity write SetQuantity;
    property MetricLength: TLaMetricLength read GetMetricLength;
    property UsLength: TLaUsLength read GetUsLength;
    property MetricTemperature: TLaMetricTemperature read GetMetricTemperature;
    property UsTemperature: TLaUsTemperature read GetUsTemperature;
    property MetricValue: Double read GetMetricValue write SetMetricValue;
    property UsValue: Double read GetUsValue write SetUsValue;
  end;

var
  MainForm: TMainForm;

implementation

{$R *.fmx}

function TMainForm.GetQuantity: TQuantity;
begin
  Result := TQuantity(TypeCombo.ListBox.Items.Objects[TypeCombo.ItemIndex]);
end;

procedure TMainForm.SetQuantity(value: TQuantity);
begin
  TypeCombo.ItemIndex := Ord(value);
  TypeComboChange(Self);
end;

function TMainForm.GetMetricLength: TLaMetricLength;
begin
  Result := TLaMetricLength(MetricCombo.ListBox.Items.Objects[MetricCombo.ItemIndex]);
end;

function TMainForm.GetUsLength: TLaUsLength;
begin
  Result := TLaUsLength(UsCombo.ListBox.Items.Objects[UsCombo.ItemIndex]);
end;

function TMainForm.GetMetricTemperature: TLaMetricTemperature;
begin
  Result := TLaMetricTemperature(MetricCombo.ListBox.Items.Objects[MetricCombo.ItemIndex]);
end;

function TMainForm.GetUsTemperature: TLaUsTemperature;
begin
  Result := TLaUsTemperature(UsCombo.ListBox.Items.Objects[UsCombo.ItemIndex]);
end;

function TMainForm.GetMetricValue: Double;
begin
  try
    Result := StrToFloat(MetricEdit.Text);
  except
    Result := 0;
  end;
end;

procedure TMainForm.SetMetricValue(value: Double);
begin
  FUpdating := True;
  try
    MetricEdit.Text := FloatToStrF(value, ffNumber, 10, 2);
  finally
    FUpdating := False;
  end;
end;

function TMainForm.GetUsValue: Double;
begin
  try
    Result := StrToFloat(UsEdit.Text);
  except
    Result := 0;
  end;
end;

procedure TMainForm.SetUsValue(value: Double);
begin
  FUpdating := True;
  try
    UsEdit.Text := FloatToStrF(value, ffNumber, 10, 2);
  finally
    FUpdating := False;
  end;
end;

procedure TMainForm.FormCreate(Sender: TObject);

  function GetQuantityName(value: TQuantity): String;
  resourcestring
    SLength = 'Length';
    STemperature = 'Temperature';
  begin
    case value of
      qaLength: Result := SLength;
      qaTemperature: Result := STemperature;
    end;
  end;

var
  q: TQuantity;
begin
  // Populate type combo
  for q := Low(q) to High(q) do
    TypeCombo.ListBox.Items.AddObject(GetQuantityName(q), TObject(q));

  Quantity := Low(TQuantity);
end;

procedure TMainForm.TypeComboChange(Sender: TObject);

  procedure AddMetric(const str: String; value: Integer);
  begin
    MetricCombo.ListBox.Items.AddObject(str, TObject(value));
  end;

  procedure AddUs(const str: String; value: Integer);
  begin
    UsCombo.ListBox.Items.AddObject(str, TObject(value));
  end;

  function GetMetricLength(value: TLaMetricLength): String;
  resourcestring
    SKilometers = 'Kilometers';
    SMeters = 'Meters';
    SCentimeters = 'Centimeters';
    SMillimeters = 'Millimeters';
  begin
    case value of
      mlKm: Result := SKilometers;
      mlCm: Result := SCentimeters;
      mlMm: Result := SMillimeters;
    else
      Result := SMeters;
    end;
  end;

  function GetUsLength(value: TLaUsLength): String;
  resourcestring
    SMiles = 'Miles';
    SYards = 'Yards';
    SFeet = 'Feet';
    SInches = 'Inches';
  begin
    case value of
      ulMile: Result := SMiles;
      ulFoot: Result := SFeet;
      ulInch: Result := SInches;
    else
      Result := SYards;
    end;
  end;

  function GetMetricTemperature(value: TLaMetricTemperature): String;
  resourcestring
    SCelsius = 'Celsius';
    SKelvin = 'Kelvin';
  begin
    case value of
      mtKelvin: Result := SKelvin;
    else
      Result := SCelsius;
    end;
  end;

  function GetUsTemperature(value: TLaUsTemperature): String;
  resourcestring
    SFahrenheit = 'Fahrenheit';
  begin
    Result := SFahrenheit;
  end;

var
  ml: TLaMetricLength;
  ul: TLaUsLength;
  mt: TLaMetricTemperature;
  ut: TLaUsTemperature;
begin
  MetricCombo.ListBox.Items.Clear;
  UsCombo.ListBox.Items.Clear;

  case Quantity of
    qaLength:
    begin
      for ml := Low(ml) to High(ml) do
        AddMetric(GetMetricLength(ml), Ord(ml));

      for ul := Low(ul) to High(ul) do
        AddUs(GetUsLength(ul), Ord(ul));
    end;

    qaTemperature:
    begin
      for mt := Low(mt) to High(mt) do
        AddMetric(GetMetricTemperature(mt), Ord(mt));

      for ut := Low(ut) to High(ut) do
        AddUs(GetUsTemperature(ut), Ord(ut));
    end;
  end;

  FUpdating := True;
  try
    MetricCombo.ItemIndex := 0;
    UsCombo.ItemIndex := 0;
  finally
    FUpdating := False;
  end;

  MetricEditChange(Self);
end;

procedure TMainForm.MetricEditChange(Sender: TObject);
begin
  if FUpdating or (MetricEdit.Text = '') then
    Exit;

  case Quantity of
    qaLength:
      UsValue := MetricLengthToUsLength(MetricValue, MetricLength, UsLength);

    qaTemperature:
      UsValue := MetricTemperatureToUsTemperature(MetricValue, MetricTemperature, UsTemperature);
  end;
end;

procedure TMainForm.UsEditChange(Sender: TObject);
begin
  if FUpdating or (UsEdit.Text = '') then
    Exit;

  case Quantity of
    qaLength:
      MetricValue := UsLengthToMetricLength(UsValue, UsLength, MetricLength);

    qaTemperature:
      MetricValue := UsTemperatureToMetricTemperature(UsValue, UsTemperature, MetricTemperature);
  end;
end;

procedure TMainForm.ExitMenuClick(Sender: TObject);
begin
  Close;
end;

procedure TMainForm.AboutMenuClick(Sender: TObject);
resourcestring
  SAboutConverter = 'About Converter';
  SAbout = 'Converter is an application that converts metric units to US units and back.';
begin
  ShowMessage(SAboutConverter + #13#10#13#10 + SAbout);
end;

end.
