/*
  Copyrights 2005-2011 Sisulizer

  A class that contains functions to access resources and satellite assemblies.
  This code is common for Windows Forms and WPF.
*/

using System;
using System.Collections;
using System.Globalization;
using System.IO;
using System.Reflection;
using System.Resources;
using System.Threading;

namespace Sisulizer
{
  public class Resources
  {
    static private Assembly assembly = null;
    static private Assembly mainAssembly = null;
    static private string assemblyName = "";
    static private string culture = "";

    // Combines assembly name and resource name to get full resource name
    static private string GetResourceName(string resourceName)
    {
      // MyImage.png -> MyAssembly.MyImage.png
      return assemblyName + '.' + resourceName;
    }

    // Combines assembly name and culture specific resource name to get full culture specific resource name
    static private string GetCultureResourceName(string resourceName)
    {
      // MyImage.png -> MyAssembly.MyImage.fi.png or MyAssembly.MyImage.fi-FI.png
      int p = resourceName.LastIndexOf('.');

      if (p >= 0)
        resourceName = resourceName.Substring(0, p + 1) + culture + resourceName.Substring(p, resourceName.Length - p);

      return GetResourceName(resourceName);
    }

    // Return culture specific file name
    static private string GetCultureFileName(string fileName)
    {
      // MyImage.png -> fi\MyImage.png
      string dir = Path.GetDirectoryName(fileName);

      if (dir == "")
        dir = Path.GetDirectoryName(mainAssembly.Location);

      return dir + "\\" + culture + "\\" + Path.GetFileName(fileName);
    }

    // Initialize assembly that contains resources. 
    // This must be called before calling GetResource or GetStandardResource
    static public void Initialize(Assembly mainAssembly)
    {
      Sisulizer.Resources.mainAssembly = mainAssembly;

      // Get assembly name
      assemblyName = mainAssembly.FullName;

      int p = assemblyName.IndexOf(',');

      if (p >= 0)
      {
        assemblyName = assemblyName.Substring(0, p);
      }

      // Get the active locale
      culture = Thread.CurrentThread.CurrentUICulture.Name;

      // Try to load the satellite assembly file
      // Try first exact culture info match (e.g. fi-FI)
      try
      {
        assembly = mainAssembly.GetSatelliteAssembly(new CultureInfo(culture));
      }
      catch (Exception)
      {
      }

      // If not found try language only match (e.g. fi)
      if (assembly == null)
      {
        p = culture.IndexOf('-');

        if (p >= 0)
        {
          culture = culture.Substring(0, p);

          try
          {
            assembly = mainAssembly.GetSatelliteAssembly(new CultureInfo(culture));
          }
          catch (Exception)
          {
          }
        }
      }

      // No satellite assembly file found. Use the main assembly.
      if (assembly == null)
        assembly = mainAssembly;
    }

    // Return stream to a given resource
    static public Stream GetResource(string resourceName)
    {
      if (assembly == null)
        throw new Exception("Sisulizer.Resources.Initialize is not called");

      if (assembly == mainAssembly)
      {
        // No satellite assembly file
        // MyApplication.exe: MyApplication.MyImage.png
        return assembly.GetManifestResourceStream(GetResourceName(resourceName));
      }
      else
      {
        // Use satellite assembly file

        // First try culture specific resource name in the satellite assembly file
        // fi\MyApplication.resources.dll: MyApplication.MyImage.fi.png
        Stream value = assembly.GetManifestResourceStream(GetCultureResourceName(resourceName));

        if (value != null)
          return value;

        // Not found, try original resource name in the satellite assembly file
        // fi\MyApplication.resources.dll: MyApplication.MyImage.png
        value = assembly.GetManifestResourceStream(GetResourceName(resourceName));

        if (value != null)
          return value;

        // Not found, try original resource
        // MyApplication.exe: MyApplication.MyImage.png
        if (mainAssembly != null)
          return mainAssembly.GetManifestResourceStream(GetResourceName(resourceName));
        else
          return null;
      }
    }

    // Return stream to a given standard resource
    static public Stream GetStandardResource(string resourceName)
    {
      Stream resourceStream = GetResource("g.resources");

      using (ResourceReader resourceReader = new ResourceReader(resourceStream))
      {
        foreach (DictionaryEntry resourceEntry in resourceReader)
        {
          if (resourceEntry.Key.ToString() == resourceName)
          {
            return resourceEntry.Value as Stream;
          }
        }
      }

      return null;
    }

    // Return stream to a given file
    static public Stream GetFileResource(string fileName)
    {
      string thisFileName = GetCultureFileName(fileName);

      if (!File.Exists(thisFileName))
        thisFileName = fileName;

      return new FileStream(thisFileName, FileMode.Open, FileAccess.Read);
    }

    // Return array that contains the languages satellite assemblies of the current application
    static public CultureInfo[] GetAvailableLanguages()
    {
      return GetAvailableLanguages(Assembly.GetExecutingAssembly().Location);
    }

    // Check if there exists any satellite assembly files
    static public void CheckSatelliteAssemblyCount()
    {
      if (GetAvailableLanguages().Length == 0)
        throw new Exception("In order to change language you must first create satellite assembly files.");
    }

    // Return array that contains the languages satellite assemblies of the given assembly file
    static public CultureInfo[] GetAvailableLanguages(string fileName)
    {
      string dir = new FileInfo(fileName).Directory.FullName;
      DirectoryInfo di = new DirectoryInfo(dir);
      DirectoryInfo[] dirs = di.GetDirectories();
      CultureInfo[] cultures = CultureInfo.GetCultures(CultureTypes.AllCultures);
      ArrayList assemblies = new ArrayList();

      for (int i = 0; i < dirs.Length; i++)
      {
        di = dirs[i];
        string str = di.Name;
        bool found = false;

        for (int j = 0; j < cultures.Length; j++)
        {
          if ((cultures[j].Name == str) || (cultures[j].TwoLetterISOLanguageName == str))
          {
            found = true;
            break;
          }
        }

        if (found)
          assemblies.Add(new CultureInfo(str));
      }

      CultureInfo[] languages = new CultureInfo[assemblies.Count];

      for (int i = 0; i < assemblies.Count; i++)
        languages[i] = assemblies[i] as CultureInfo;

      return languages;
    }
  }
}