{
  Copyright 2005-2011 Sisulizer

  Unit that contains I18N functions for Windows VCL and .NET VCL.
  This compiles with Delphi 5 and later.
}

{$I LaDef.inc}

unit LaCommon;

interface

const
  LOCALE_SEPARATOR = '_';

  MILE_AS_METERS = 1609.3;
  YARD_AS_METERS = 0.9144;

  INCH_IN_FEET = 12;
  FEET_IN_YARD = 3;
  YARDS_IN_MILE = 1760;

  CELSIUS_OFFSET = 273.15;
  FAHRENHEIT_OFFSET = 32;
  FAHRENHEIT_FACTOR = 1.8;

type
  TLaLocaleType =
  (
    ltUser,
    ltSystem
  );

  TQuantity =
  (
    qaLength,
    qaTemperature
  );

  TLaMeasurementSystem =
  (
    msMetric,
    msUS
  );

  TLaMetricLength =
  (
    mlKm,
    mlM,
    mlCm,
    mlMm
  );

  TLaUsLength =
  (
    ulMile,
    ulYard,
    ulFoot,
    ulInch
  );

  TLaMetricTemperature =
  (
    mtCelsius,
    mtKelvin
  );

  TLaUsTemperature =
  (
    utFahrenheit
  );

{$IFNDEF UNICODE_VCL}
  UnicodeString = WideString;
  RawByteString = AnsiString;
{$ENDIF}


function MetricLengthToUsLength(
  value: Double;
  metricLength: TLaMetricLength;
  usLength: TLaUsLength): Double;

function UsLengthToMetricLength(
  value: Double;
  usLength: TLaUsLength;
  metricLength: TLaMetricLength): Double;

function MetricTemperatureToUsTemperature(
  value: Double;
  metricTemperature: TLaMetricTemperature;
  usTemperature: TLaUsTemperature): Double;

function UsTemperatureToMetricTemperature(
  value: Double;
  usTemperature: TLaUsTemperature;
  metricTemperature: TLaMetricTemperature): Double;

procedure ParseLocaleString(
  value: String;
  var language, country, variant: String);

function GetLocaleParameter(const optionChars: String = '-/ '): String;
{$IFDEF CIL}
procedure SetLocaleParameter(const optionChars: String = '-/ ');
{$ENDIF}

function MakeLangId(primaryLanguage, subLanguage: Integer): Integer;
function LocaleToPrimary(locale: Integer): Integer;
function LocaleToSub(locale: Integer): Integer;

{$IFNDEF CIL}
function LocaleToExtension(locale: Integer): String;
function LocaleToIso(locale: Integer): String;
function IsCurrentLocaleBidi: Boolean;
function IsCurrentLocaleAsian: Boolean;

function AnsiToUnicode(const str: RawByteString; codePage: Integer = 0): UnicodeString;
function UnicodeToAnsi(const str: UnicodeString; codePage: Integer = 0): RawByteString;
{$ENDIF}

function GetCurrentLocale: Integer;

function IsLocaleAsian(value: String): Boolean;
function IsLocaleBidi(value: String): Boolean;

function GetMeasurementSystem(locale: Integer = 0): TLaMeasurementSystem;

var
  LocaleType: TLaLocaleType;
  MeasurementSystem: TLaMeasurementSystem;

implementation

uses
{$IFNDEF CIL}
  Windows,
{$ELSE}
  System.Threading,
  System.Globalization,
{$ENDIF}
  SysUtils;

function MetricLengthToUsLength(
  value: Double;
  metricLength: TLaMetricLength;
  usLength: TLaUsLength): Double;
var
  meters, yards: Double;
begin
  case metricLength of
    mlKm: meters := 1000*value;
    mlCm: meters := value/100;
    mlMm: meters := value/1000;
  else
    meters := value;
  end;

  yards := meters/YARD_AS_METERS;

  case usLength of
    ulMile: Result := yards/YARDS_IN_MILE;
    ulFoot: Result := FEET_IN_YARD*yards;
    ulInch: Result := FEET_IN_YARD*INCH_IN_FEET*yards;
  else
    Result := yards;
  end;
end;

function UsLengthToMetricLength(
  value: Double;
  usLength: TLaUsLength;
  metricLength: TLaMetricLength): Double;
var
  meters, yards: Double;
begin
  case usLength of
    ulMile: yards := YARDS_IN_MILE*value;
    ulFoot: yards := value/FEET_IN_YARD;
    ulInch: yards := value/(FEET_IN_YARD*INCH_IN_FEET);
  else
    yards := value;
  end;

  meters := YARD_AS_METERS*yards;

  case metricLength of
    mlKm: Result := meters/1000;
    mlCm: Result := 100*meters;
    mlMm: Result := 1000*meters;
  else
    Result := meters;
  end;
end;

function MetricTemperatureToUsTemperature(
  value: Double;
  metricTemperature: TLaMetricTemperature;
  usTemperature: TLaUsTemperature): Double;
var
  celsius, fahrenheit: Double;
begin
  case metricTemperature of
    mtKelvin: celsius := value - CELSIUS_OFFSET;
  else
    celsius := value;
  end;

  fahrenheit := FAHRENHEIT_FACTOR*celsius + FAHRENHEIT_OFFSET;

  Result := fahrenheit;
end;

function UsTemperatureToMetricTemperature(
  value: Double;
  usTemperature: TLaUsTemperature;
  metricTemperature: TLaMetricTemperature): Double;
var
  celsius, fahrenheit: Double;
begin
  fahrenheit := value;

  celsius := (fahrenheit - FAHRENHEIT_OFFSET)/FAHRENHEIT_FACTOR;

  case metricTemperature of
    mtKelvin: Result := celsius + CELSIUS_OFFSET;
  else
    Result := celsius;
  end;
end;

procedure ParseLocaleString(
  value: String;
  var language, country, variant: String);

  function GetString: String;
  var
    p: Integer;
  begin
    p := Pos(LOCALE_SEPARATOR, value);

    if p > 0 then
    begin
      Result := Copy(value, 1, p - 1);
      Delete(value, 1, p);
    end
    else
    begin
      Result := value;
      value := '';
    end;
  end;

begin
  language := GetString;
  country := GetString;
  variant := GetString;
end;

function MakeLangId(primaryLanguage, subLanguage: Integer): Integer;
begin
  Result := (subLanguage shl 10) or primaryLanguage;
end;

function LocaleToPrimary(locale: Integer): Integer;
begin
  Result := locale and $3FF;
end;

function LocaleToSub(locale: Integer): Integer;
begin
  Result := (locale shr 10) and $3F;
end;

function IsLocaleBidi(value: String): Boolean;
begin
  if Length(value) > 2 then
    value := Copy(value, 1, 2);

  value := LowerCase(value);

  Result := (value = 'ar') or (value = 'he') or (value = 'fa');
end;

function IsLocaleAsian(value: String): Boolean;
begin
  if Length(value) > 2 then
    value := Copy(value, 1, 2);

  value := LowerCase(value);

  Result := (value = 'jp') or (value = 'ch') or (value = 'ko');
end;

{$IFNDEF CIL}
function IsCurrentLocaleBidi: Boolean;
begin
  Result := IsLocaleBidi(LocaleToExtension(GetCurrentLocale));
end;

function IsCurrentLocaleAsian: Boolean;
begin
  Result := IsLocaleAsian(LocaleToExtension(GetCurrentLocale));
end;

function LocaleToExtension(locale: Integer): String;
begin
  if locale = 0 then
    Result := ''
  else
  begin
    Result := GetLocaleStr(locale, LOCALE_SABBREVLANGNAME, '');

    if LocaleToSub(locale) = SUBLANG_NEUTRAL then
      Delete(Result, Length(Result), 1);
  end;
end;

function LocaleToIso(locale: Integer): String;
begin
  Result := GetLocaleStr(locale, LOCALE_SISO639LANGNAME, '');

  if LocaleToSub(locale) <> SUBLANG_NEUTRAL then
    Result := Result + '_' + GetLocaleStr(locale, LOCALE_SISO3166CTRYNAME, '');
end;

function AnsiToUnicode(const str: RawByteString; codePage: Integer): UnicodeString;
var
  len: Integer;
begin
{$IFDEF UNICODE_VCL}
  if codePage = 0 then
    codePage := StringCodePage(str);

  if codePage = 0 then
    codePage := DefaultSystemCodePage;
{$ENDIF}

  if str = '' then
    Result := ''
  else
  begin
    len := MultiByteToWideChar(codePage, 0, PAnsiChar(str), -1, nil, 0);

    if len > 1 then
    begin
      SetLength(Result, len - 1);
      MultiByteToWideChar(codePage, 0, PAnsiChar(str), -1, PWideChar(Result), len);
    end
    else
      Result := '';
  end;
end;

function UnicodeToAnsi(const str: UnicodeString; codePage: Integer): RawByteString;
var
  len: Integer;
begin
{$IFDEF UNICODE_VCL}
  if codePage = 0 then
    codePage := DefaultSystemCodePage;
{$ENDIF}

  if str = '' then
    Result := ''
  else
  begin
    len := WideCharToMultiByte(codePage, 0, PWideChar(str), -1, nil, 0, nil, nil);

    if len > 1 then
    begin
      SetLength(Result, len - 1);
      WideCharToMultiByte(codePage, 0, PWideChar(str), -1, PAnsiChar(Result), len, nil, nil);
    end
    else
      Result := '';
  end;

{$IFDEF UNICODE_VCL}
  SetCodePage(Result, codePage, False);
{$ENDIF}
end;
{$ENDIF}

function GetLocaleParameter(const optionChars: String): String;
var
  option: Char;
  i, j: Integer;
  str, language, country, variant: String;
begin
  Result := '';

  if ParamCount = 0 then
    Exit;

  for i := 1 to Length(optionChars) do
  begin
    option := optionChars[i];
    j := 0;

    while j < ParamCount do
    begin
      Inc(j);
      str := ParamStr(j);

      if option <> ' ' then
      begin
        if str[1] = option then
          Delete(str, 1, 1)
        else
          Continue;
      end;

      ParseLocaleString(str, language, country, variant);

      if (Length(language) = 2) and ((country = '') or (Length(country) = 2)) then
      begin
        Result := str;
        Break;
      end;
    end;
  end;
end;

{$IFDEF CIL}
procedure SetLocaleParameter(const optionChars: String);
var
  locale: String;
  ci: CultureInfo;
begin
  locale := GetLocaleParameter(optionChars);

  if locale <> '' then
  begin
    ci := CultureInfo.Create(locale);

    Thread.CurrentThread.CurrentUICulture := ci;

    if not ci.IsNeutralCulture then
      Thread.CurrentThread.CurrentCulture := ci;
  end
  else
    Thread.CurrentThread.CurrentUICulture := CultureInfo.CurrentCulture;
end;
{$ENDIF}

function GetCurrentLocale: Integer;
begin
{$IFNDEF CIL}
  if LocaleType = ltSystem then
    Result := GetSystemDefaultLCID
  else
    Result := GetUserDefaultLCID;
{$ELSE}
  Result := Thread.CurrentThread.CurrentUICulture.LCID;
{$ENDIF}
end;

function GetMeasurementSystem(locale: Integer): TLaMeasurementSystem;
begin
  if locale = 0 then
    locale := GetCurrentLocale;

  if locale = $0409 then
    Result := msUs
  else
    Result := msMetric;
end;

initialization
  MeasurementSystem := GetMeasurementSystem;
end.
