{
  Copyright 2005-2011 Sisulizer

  Unit that contains I18N functions for Windows VCL.
  This compiles with Delphi 5 and later.
}

{$I LaDef.inc}

unit LaResource;

interface

uses
  Classes,
  Controls,
  Graphics,
  Forms,
  StdCtrls,
  Windows,
  LaCommon;

const
{$IFDEF VCL14}
  LOCALE_OVERRIDE_KEY = 'Software\Embarcadero\Locales';
  OLD_LOCALE_OVERRIDE_KEY = 'Software\CodeGear\Locales';
{$ELSE}
  {$IFDEF UNICODE_VCL}
  LOCALE_OVERRIDE_KEY = 'Software\CodeGear\Locales';
  OLD_LOCALE_OVERRIDE_KEY = 'Software\Borland\Locales';
  {$ELSE}
  LOCALE_OVERRIDE_KEY = 'Software\Borland\Locales';
  {$ENDIF}
{$ENDIF}

  DEFAULT_SECTION = 'Sisulizer';
  DEFAULT_IDENT = 'Locale';

  CSIDL_PERSONAL = $0005;
  CSIDL_PROGRAM_FILES = $0026;
  CSIDL_PROGRAM_FILES_COMMON = $002b;

  PROCEDURE_DATA_SIZE_C = 6;

  VCL_FORM_HEADER_C: AnsiString = 'TPF0';

type
  // Specifies the day of the week
  TLaDayOfWeek =
  (
    wdMonday,
    wdTuesday,
    wdWednesday,
    wdThursday,
    wdFriday,
    wdSaturday,
    wdSunday
  );

  // Specifies what is the first week of the year
  TLaFirstWeekOfYear =
  (
    fwFirstPart,  // First week that has at least one day in January
    fwFirstFull,  // First full week of January
    fwFirst4      // First week that has at least four days in January
  );

  TLaUiLayout =
  (
    ulLeftToRight,  // The layout of user interface is Western (from left to right)
    ulRightToLeft   // The layout of user interface is Middle-Eastern (from right to left)
  );

  TLocaleFontSignature = packed record
    fsUsb: array[0..3] of DWord;
    fsCsbDefault: array[0..1] of DWord;
    fsCsbSupported: array[0..1] of DWord;
  end;
  PLocaleFontSignature = ^TLocaleFontSignature;

  TLaCharsetInfo = record
    charSet: TFontCharset;
    codePage: Integer;
  end;

  TLanguageListItem = class(TObject)
  private
    FCode: String;
    FName: UnicodeString;

    function GetCharset: TFontCharset;

  public
    constructor Create(const code: String; const name: UnicodeString);

    class procedure DrawItem(
      parent: TWinControl;
      listBox: TListBox;
      index: Integer;
      rect: TRect;
      state: TOwnerDrawState);

    property Charset: TFontCharset read GetCharset;
    property Code: String read FCode;
    property Name: UnicodeString read FName;
  end;

  // Specifies how embedded resources are extracted
  TExtractOption =
  (
    eoCheckDate,   // Checks the date of the resource files. If they are older than EXE extracts the resources files
    eoRemoveFiles  // Removes extracted file when applications quits
  );
  TExtractOptions = set of TExtractOption;

  TLaLocaleSource =
  (
    lsLoaded,  // Use the locale of the resource DLL that is currently loaded
    lsSystem   // Use the locale specified in the system through Control Panel
  );

  // Specifies the options of the resource set functions
  TLaResourceOption =
  (
    roNoThreadLocale,    // Do not set the language of the current thread to match the new language
    roNoLocaleVariables, // Do not update VCL's locale variables to match the new language
    roNoUpdateBidiMode,  // Do not update Application.BiDiMode
    roSaveLocale,        // Save the current locale to VCL's system registry
    roFlipChildren       // Flip forms if needed
  );
  TLaResourceOptions = set of TLaResourceOption;

  TProcedureData = record
    Address: Pointer;
    Data: array[0..PROCEDURE_DATA_SIZE_C] of Byte;
  end;

// Updates the locale variables to match the initially loaded resource DLL
procedure CheckLocaleVariables;

procedure UpdateFormatSettings(locale: Integer);

// Return true if the give locale is ompatible to the current system settings
function IsLocaleCompatible(locale: Integer): Boolean;

function ExtensionToLocale(const value: String): Integer;
function GetLocaleExtension: String;

function LocaleToCodePage(locale: Integer): Integer;

function LocaleToIso639(locale: Integer): String;

function GetWindowsLocaleName(
  locale: Integer;
  useNative: Boolean = False): String;

function GetWindowsLanguageWideName(
  locale: Integer;
  useNative: Boolean = False): UnicodeString;

function GetWindowsLocaleWideName(
  locale: Integer;
  useNative: Boolean = False): UnicodeString;

function Iso639ToLocale(
  const language: String;
  const country: String = ''): Integer;

function IsoToLocale(const locale: String): Integer;

// Returns true if a resource DLL has been loaded.
// Return false if the application uses the orignal resource of the EXE file.
function IsResourceFileLoaded: Boolean;

// Returns the locale of the currently loaded resources.
// - defaultLocale returns this if no resource DLL is currently loaded.
function GetActiveResourceLocale(defaultLocale: String = ''): String;

// Returns true of active locale is Asian
function IsActiveLocaleAsian: Boolean;

// Returns true of active locale is bi-directional
function IsActiveLocaleBidi: Boolean;

// Returns true of previous locale is bi-directional
function IsPreviousLocaleBidi: Boolean;

procedure InitializeForm(form: TCustomForm);

// Call this function in the initialization section of your main form
procedure InitializeApplication(layout: TLaUiLayout = ulLeftToRight);

// Returns a list of available resource DLL languages.
// - languages returns a list of available languages
// - exeFileName specifies the name of the application. If empty the current application is used.
// - compatibleOnly if true only those resource DLL languages are added that are compatible to the current system settings.
function GetAvailableResources(
  languages: TList;
  exeFileName: String = '';
  compatibleOnly: Boolean = False): Integer; overload;

function GetAvailableResources(
  languages: TStrings;
  exeFileName: String = '';
  compatibleOnly: Boolean = False): Integer; overload;

// Returns the number of the available resource DLLs
// For parameters see GetAvailableResources
function GetAvailableResourceCount(
  exeFileName: String = '';
  compatibleOnly: Boolean = False): Integer;

// Checks the resource DLL count. If this is 0 raises an exception.
procedure CheckResourceCount;

// Extracts embedded resource DLLs from EXE to the application directory
// - options       specifies the extract options
// - resourceName  specifies the resource type name of the resource DLLs. If empty SISULIZER is used.
function ExtractResourceFiles(
  options: TExtractOptions = [eoCheckDate];
  resourceName: String = ''): Integer;

// Set a new resource DLL file and initializes the language depend values
// - code specifies the file extension without period of the new resource DLL.
// - updateThreadLocale if true set the language of the current thread to match the new language
// - updateLocaleVariables if true update VCL's locale variables to match the new language
function SetNewResourceFile(
  const code: String = '';
  options: TLaResourceOptions = [];
  const originalCode: String = ''): Boolean; overload;

function SetNewResourceFile(
  locale: Integer;
  options: TLaResourceOptions = [];
  originalLocale: Integer = 0): Boolean; overload;

procedure DisableResourceDllUse;

// Load a new resource DLL file. If you call SetNewResourceFile this should not be called.
function LoadNewResourceFile(const code: String = ''): Integer;

procedure SetLocaleParameter(const optionChars: String = '-/ ');

function GetLanguageFileExtension(locale: Integer): String;
function GetLanguageFile(const exeFileName: String; const code: String): String;

function DoesResourceFileVersionMatch: Boolean; overload;
function DoesResourceFileVersionMatch(resInstance: LongWord): Boolean; overload;
function DoesLocaleResourceFileVersionMatch(const code: String): Boolean;

procedure SetCurrentDefaultLocaleReg;
function GetCurrentDefaultLocaleReg: String;
function ClearCurrentDefaultLocaleReg: Boolean;

procedure SetDefaultLocaleReg(const fileName: String; const code: String);
function GetDefaultLocaleReg(const fileName: String): String; overload;
function GetDefaultLocaleReg(const fileName: String; out keyExists: Boolean): String; overload;
function ClearDefaultLocaleReg(const fileName: String): Boolean;

function IsNtOrLater: Boolean;
function HasWriteAccess(const dir: String): Boolean;
function GetFolderPath(nFolder: Integer): String;

function GetCharSetInfo(langId: Integer): TLaCharsetInfo;
function GetActiveFontCharset: TFontCharset;
function GetActiveCodePage: Integer;
function GetActiveLangId(def: Integer): Integer;

function GetResourceInstance: LongWord;
function GetFormLocale: Integer;
function GetResourceLocale(resType, resName: PChar; instance: THandle = 0): Integer;
procedure InitializeStaticLocale(default: Integer);

{$IFDEF VCL6}
  {$IFDEF WIN32}
procedure RemapProcedure(oldProc, newProc: Pointer; var data: TProcedureData);
procedure RestoreProcedure(originalProc: Pointer; data: TProcedureData);
  {$ENDIF}
{$ENDIF}

var
  FirstDayOfWeek: TLaDayOfWeek;
  FirstWeekOfYear: TLaFirstWeekOfYear;
  LaResourceOptions: TLaResourceOptions;
  LoadedResourceLocale: String;
  PreviouslyLoadedResourceLocale: String;
  UiLayout: TLaUiLayout;

implementation

uses
{$IFNDEF VCL6}
  FileCtrl,
{$ENDIF}
  IniFiles,
  Registry,
  SysUtils,
  LaTranslator;

const
  APPLICATION_RESOURCE = 'SISULIZER';
  APPLICATION_DIR = 'Sisulizer';

var
  FCountry: String;
  FExtension: String;
  FExtractedResourceFiles: TStringList;
  FFileDate: TDateTime;
  FFileName: String;
  FFormResourceLocale: Integer;
  FLanguage: String;
  FLocale: Integer;
  FOptions: TExtractOptions;
  FResourceFileCount: Integer;

type
  TSHGetFolderPath = function(hwndOwner: HWnd; nFolder: Integer; hToken: THandle; dwFlags: DWord; lpszPath: PChar): HResult; stdcall;

function EnumEraNames(names: PChar): Integer; stdcall;
var
  i: Integer;
begin
  Result := 0;

  i := Low(EraNames);

  while EraNames[I] <> '' do
  begin
    if (i = High(EraNames)) then
      Exit
    else
      Inc(i);
  end;

  EraNames[i] := names;
  Result := 1;
end;

function EnumEraYearOffsets(yearOffsets: PChar): Integer; stdcall;
var
  i: Integer;
begin
  Result := 0;

  i := Low(EraYearOffsets);

  while EraYearOffsets[i] <> -1 do
  begin
    if (i = High(EraYearOffsets)) then
      Exit
    else
      Inc(I);
  end;

  EraYearOffsets[i] := StrToIntDef(yearOffsets, 0);
  Result := 1;
end;

function LocaleToCodePage(locale: Integer): Integer;
begin
  if locale = LANG_CHINESE then
    Result := 936
  else
    Result := StrToInt(GetLocaleStr(locale, LOCALE_IDEFAULTANSICODEPAGE, '0'));
end;

function IsLocaleCompatible(locale: Integer): Boolean;
begin
{$IFDEF UNICODE_VCL}
  Result := True;
{$ELSE}
  Result := LocaleToCodePage(locale) = Integer(GetACP);
{$ENDIF}
end;

function LocaleToIso639(locale: Integer): String;
begin
  Result := GetLocaleStr(locale, LOCALE_SISO639LANGNAME, '');

  if LocaleToSub(locale) <> SUBLANG_NEUTRAL then
    Result := Result + LOCALE_SEPARATOR + GetLocaleStr(locale, LOCALE_SISO3166CTRYNAME, '');
end;

function GetLocaleExtension: String;
begin
  Result := LOCALE_SEPARATOR + LocaleToIso639(GetCurrentLocale);
end;

{$IFNDEF UNICODE_VCL}
function CharInSet(c: AnsiChar; const charSet: TSysCharSet): Boolean;
begin
  Result := c in charSet;
end;
{$ENDIF}

function TranslateDateFormat(
  locale: Integer;
  const formatStr: String): String;
var
  i: Integer;
{$IFDEF VCL6}
  l: Integer;
{$ENDIF}
  calendarType: CALTYPE;
begin
  i := 1;
  Result := '';

  calendarType := StrToIntDef(GetLocaleStr(locale, LOCALE_ICALENDARTYPE, '1'), 1);

  if not (calendarType in [CAL_JAPAN, CAL_TAIWAN, CAL_KOREA]) then
  begin
    if SysLocale.PriLangID in [LANG_JAPANESE, LANG_CHINESE, LANG_KOREAN] then
    begin
      while i <= Length(formatStr) do
      begin
        if not CharInSet(formatStr[i], ['g', 'G']) then
          Result := Result + formatStr[i];

        Inc(i);
      end;
    end
    else
      Result := formatStr;

    Exit;
  end;

  while i <= Length(formatStr) do
  begin
    if CharInSet(formatStr[i], LeadBytes) then
    begin
{$IFDEF VCL6}
      l := CharLength(formatStr, i);
      Result := Result + Copy(formatStr, i, l);
      Inc(i, l);
{$ELSE}
      Result := Result + Copy(formatStr, i, 2);
      Inc(i, 2);
{$ENDIF}
    end
    else
    begin
      if StrLIComp(@formatStr[i], 'gg', 2) = 0 then
      begin
        Result := Result + 'ggg';
        Inc(i, 1);
      end
      else if StrLIComp(@formatStr[i], 'yyyy', 4) = 0 then
      begin
        Result := Result + 'eeee';
        Inc(i, 4 - 1);
      end
      else if StrLIComp(@formatStr[i], 'yy', 2) = 0 then
      begin
        Result := Result + 'ee';
        Inc(i, 2 - 1);
      end
      else if CharInSet(formatStr[i], ['y', 'Y']) then
        Result := Result + 'e'
      else
        Result := Result + formatStr[i];

      Inc(i);
    end;
  end;
end;

{$IFNDEF VCL10}
function LcidToCodePage(locale: LCID): Integer;
var
  buffer: array [0..6] of Char;
begin
  GetLocaleInfo(locale, LOCALE_IDEFAULTANSICODEPAGE, buffer, SizeOf(buffer));
  Result:= StrToIntDef(buffer, GetACP);
end;

procedure InitSysLocale(locale: Integer);
var
  i: Integer;
  defaultLangID: LANGID;
  ansiCPInfo: TCPInfo;
  pcharA: PAnsiChar;
  bufferA: array[128..255] of Char;
  bufferW: array[128..256] of Word;

  procedure InitLeadBytes;
  var
    i: Integer;
    j: Byte;
  begin
    GetCPInfo(LcidToCodePage(SysLocale.DefaultLCID), ansiCPInfo);

    with ansiCPInfo do
    begin
      i := 0;
      while (i < MAX_LEADBYTES) and ((LeadByte[i] or LeadByte[i + 1]) <> 0) do
      begin
        for j := LeadByte[i] to LeadByte[i + 1] do
          Include(LeadBytes, Char(j));
        Inc(i, 2);
      end;
    end;
  end;

  function IsWesternGroup: Boolean;
  type
    TLanguage = $00..$1D;
  const
    NEUTRAL = TLanguage($00);
    DANISH = TLanguage($06);
    DUTCH = TLanguage($13);
    ENGLISH  = TLanguage($09);
    FINNISH = TLanguage($0B);
    FRENCH = TLanguage($0C);
    GERMAN = TLanguage($07);
    ITALIAN  = TLanguage($10);
    NORWEGIAN = TLanguage($14);
    PORTUGUSE = TLanguage($16);
    SPANISH  = TLanguage($0A);
    SWEDISH  = TLanguage($1D);
  begin
    Result := SysLocale.PriLangID in
    [
      NEUTRAL,
      DANISH,
      DUTCH,
      ENGLISH,
      FINNISH,
      FRENCH,
      GERMAN,
      ITALIAN,
      NORWEGIAN,
      PORTUGUSE,
      SPANISH,
      SWEDISH
    ];
  end;

begin
  SysLocale.DefaultLCID := $0409;
  SysLocale.PriLangID := LANG_ENGLISH;
  SysLocale.SubLangID := SUBLANG_ENGLISH_US;

  if locale <> 0 then
    SysLocale.DefaultLCID := locale;

  defaultLangID := Word(locale);

  if defaultLangID <> 0 then
  begin
    SysLocale.PriLangID := defaultLangID and $03FF;
    SysLocale.SubLangID := defaultLangID shr 10;
  end;

  LeadBytes := [];

  if Win32Platform = VER_PLATFORM_WIN32_NT then
  begin
    if IsWesternGroup then
    begin
      SysLocale.MiddleEast := False;
      SysLocale.FarEast := False;
    end
    else
    begin
      InitLeadBytes;
      SysLocale.FarEast := LeadBytes <> [];

      if SysLocale.FarEast then
      begin
        SysLocale.MiddleEast := False;
        Exit;
      end;

      for i := Low(bufferA) to High(bufferA) do
        bufferA[I] := Char(i);

      pcharA := @bufferA;

      GetStringTypeExA(
        SysLocale.DefaultLCID,
        CT_CTYPE2,
        pcharA,
        High(bufferA) - Low(bufferA) + 1,
        bufferW);

      for i := Low(bufferA) to High(bufferA) do
      begin
        SysLocale.MiddleEast := bufferW[i] = C2_RIGHTTOLEFT;

        if SysLocale.MiddleEast then
          Exit;
      end;
    end;
  end
  else
  begin
    SysLocale.MiddleEast := GetSystemMetrics(SM_MIDEASTENABLED) <> 0;
    SysLocale.FarEast := GetSystemMetrics(SM_DBCSENABLED) <> 0;

    if SysLocale.FarEast then
      InitLeadBytes;
  end;
end;
{$ENDIF}

procedure CheckLocaleVariables;
begin
  if LoadedResourceLocale <> '' then
    UpdateFormatSettings(ExtensionToLocale(LoadedResourceLocale));
end;

procedure UpdateFormatSettings(locale: Integer);

  function LocalGetLocaleStr(localeType: Integer; const default: String): String;
  begin
    Result := GetLocaleStr(locale, localeType, default);
  end;

var
  i, j, day: Integer;
  hourFormat, timePrefix, timePostfix: String;
  calendarType: CALTYPE;
begin
  if locale = LANG_CHINESE then
    locale := MakeLangId(LANG_CHINESE, SUBLANG_CHINESE_SIMPLIFIED);

{$IFNDEF VCL10}
  InitSysLocale(locale);
{$ENDIF}

  if SysLocale.FarEast then
  begin
    calendarType := StrToIntDef(GetLocaleStr(
      locale,
      LOCALE_IOPTIONALCALENDAR, '1'),
      1);

    if calendarType in [CAL_JAPAN, CAL_TAIWAN, CAL_KOREA] then
    begin
      EnumCalendarInfo(@EnumEraNames, locale, calendarType, CAL_SERASTRING);

      for j := Low(EraYearOffsets) to High(EraYearOffsets) do
        EraYearOffsets[j] := -1;

      EnumCalendarInfo(
        @EnumEraYearOffsets,
        locale,
        CalendarType,
        CAL_IYEAROFFSETRANGE);
    end;
  end;

  for i := 1 to 12 do
  begin
    {$IFDEF VCL14}FormatSettings.{$ENDIF}ShortMonthNames[i] := LocalGetLocaleStr(
      LOCALE_SABBREVMONTHNAME1 + i - 1,
      {$IFDEF VCL14}FormatSettings.{$ENDIF}ShortMonthNames[i]);

    {$IFDEF VCL14}FormatSettings.{$ENDIF}LongMonthNames[i] := LocalGetLocaleStr(
      LOCALE_SMONTHNAME1 + i - 1,
      {$IFDEF VCL14}FormatSettings.{$ENDIF}LongMonthNames[i]);
  end;

  for i := 1 to 7 do
  begin
    day := (I + 5) mod 7;

    {$IFDEF VCL14}FormatSettings.{$ENDIF}ShortDayNames[i] := LocalGetLocaleStr(
      LOCALE_SABBREVDAYNAME1 + day,
      {$IFDEF VCL14}FormatSettings.{$ENDIF}ShortDayNames[i]);

    {$IFDEF VCL14}FormatSettings.{$ENDIF}LongDayNames[i] := LocalGetLocaleStr(
      LOCALE_SDAYNAME1 + day,
      {$IFDEF VCL14}FormatSettings.{$ENDIF}LongDayNames[i]);
  end;

  {$IFDEF VCL14}FormatSettings.{$ENDIF}CurrencyString := GetLocaleStr(locale, LOCALE_SCURRENCY, '');
  {$IFDEF VCL14}FormatSettings.{$ENDIF}CurrencyFormat := StrToIntDef(GetLocaleStr(locale, LOCALE_ICURRENCY, '0'), 0);
  {$IFDEF VCL14}FormatSettings.{$ENDIF}NegCurrFormat := StrToIntDef(GetLocaleStr(locale, LOCALE_INEGCURR, '0'), 0);
  {$IFDEF VCL14}FormatSettings.{$ENDIF}ThousandSeparator := GetLocaleChar(locale, LOCALE_STHOUSAND, ',');
  {$IFDEF VCL14}FormatSettings.{$ENDIF}DecimalSeparator := GetLocaleChar(locale, LOCALE_SDECIMAL, '.');
  {$IFDEF VCL14}FormatSettings.{$ENDIF}CurrencyDecimals := StrToIntDef(GetLocaleStr(locale, LOCALE_ICURRDIGITS, '0'), 0);
  {$IFDEF VCL14}FormatSettings.{$ENDIF}DateSeparator := GetLocaleChar(locale, LOCALE_SDATE, '/');

  {$IFDEF VCL14}FormatSettings.{$ENDIF}ShortDateFormat := TranslateDateFormat(
    locale,
    GetLocaleStr(locale, LOCALE_SSHORTDATE, 'm/d/yy'));

  {$IFDEF VCL14}FormatSettings.{$ENDIF}LongDateFormat := TranslateDateFormat(
    locale,
    GetLocaleStr(locale, LOCALE_SLONGDATE, 'mmmm d, yyyy'));

  {$IFDEF VCL14}FormatSettings.{$ENDIF}TimeSeparator := GetLocaleChar(locale, LOCALE_STIME, ':');
  {$IFDEF VCL14}FormatSettings.{$ENDIF}TimeAMString := GetLocaleStr(locale, LOCALE_S1159, 'am');
  {$IFDEF VCL14}FormatSettings.{$ENDIF}TimePMString := GetLocaleStr(locale, LOCALE_S2359, 'pm');
  TimePrefix := '';
  TimePostfix := '';

  if StrToIntDef(GetLocaleStr(locale, LOCALE_ITLZERO, '0'), 0) = 0 then
    HourFormat := 'h'
  else
    HourFormat := 'hh';

  if StrToIntDef(GetLocaleStr(locale, LOCALE_ITIME, '0'), 0) = 0 then
  begin
    if StrToIntDef(GetLocaleStr(locale, LOCALE_ITIMEMARKPOSN, '0'), 0) = 0 then
      TimePostfix := ' AMPM'
    else
      TimePrefix := 'AMPM ';
  end;

  {$IFDEF VCL14}FormatSettings.{$ENDIF}ShortTimeFormat := TimePrefix + HourFormat + ':mm' + TimePostfix;
  {$IFDEF VCL14}FormatSettings.{$ENDIF}LongTimeFormat := TimePrefix + HourFormat + ':mm:ss' + TimePostfix;
  {$IFDEF VCL14}FormatSettings.{$ENDIF}ListSeparator := GetLocaleChar(locale, LOCALE_SLIST, ',');

  MeasurementSystem := GetMeasurementSystem(locale);

  FirstDayOfWeek := TLaDayOfWeek(StrToInt(GetLocaleStr(
    locale,
    LOCALE_IFIRSTDAYOFWEEK,
    '0')));

  FirstWeekOfYear := TLaFirstWeekOfYear(StrToInt(GetLocaleStr(
    locale,
    LOCALE_IFIRSTWEEKOFYEAR,
    '0')));
end;

var
  thisLocale: Integer;
  thisExt: String;

function ExtensionEnumLocales(localeStr: PChar): Integer; stdcall;
var
  id: Integer;
  str, ext, languageOnlyExt: String;
begin
  str := localeStr;
  id := StrToInt('$' + str);
  Result := 1;

  ext := GetLocaleStr(id, LOCALE_SABBREVLANGNAME, '');
  languageOnlyExt := ext;
  Delete(languageOnlyExt, Length(languageOnlyExt), 1);

  if CompareText(ext, thisExt) = 0 then
  begin
    thisLocale := id;
    Result := 0;
  end
  else if CompareText(languageOnlyExt, thisExt) = 0 then
  begin
    thisLocale := LocaleToPrimary(id);
    Result := 0;
  end;
end;

function ExtenstonToWindowsCode(const ext: String): Integer;
begin
  thisLocale := 0;
  thisExt := ext;

  EnumSystemLocales(@ExtensionEnumLocales, LCID_SUPPORTED);

  Result := thisLocale;
end;

function GetWindowsLocaleName(locale: Integer; useNative: Boolean): String;

  function LoadResString(id: Integer): String;
  var
    buffer: array[0..4095] of Char;
  begin
    SetString(
      Result,
      buffer,
      LoadString(GetResourceInstance, id, buffer, Length(buffer)))
  end;

  function Process(languageLocaleType, countryLocaleType: Integer): String;
  resourcestring
    STraditionalSort = 'Traditional Sort';
    SInternationalSort = 'International Sort';
  var
    str: String;
  begin
    Result := GetLocaleStr(locale, languageLocaleType, '');

    if LocaleToSub(locale) <> SUBLANG_NEUTRAL then
    begin
      str := GetLocaleStr(locale, countryLocaleType, '');

      if str <> '' then
      begin
        if locale = $040A then
        begin
          if useNative then
            str := str + ', Alfabetizacin tradicional'
          else
            str := str + ', ' + STraditionalSort;
        end
        else if locale = $0C0A then
        begin
          if useNative then
            str := str + ', Alfabetizacin internacional'
          else
            str := str + ', ' + SInternationalSort;
        end;

        Result := Result + ' (' + str + ')';
      end;
    end;
  end;

begin
  if useNative then
    Result := Process(LOCALE_SNATIVELANGNAME, LOCALE_SNATIVECTRYNAME)
  else
  begin
    Result := LoadResString(locale);

    if Result = '' then
      Result := Process(LOCALE_SENGLANGUAGE, LOCALE_SENGCOUNTRY);
  end;
end;

function GetLocaleStr(locale, localeType: Integer; const default: UnicodeString): UnicodeString;
var
  l: Integer;
  buffer: array[0..255] of WideChar;
begin
  l := GetLocaleInfoW(locale, localeType, buffer, SizeOf(buffer));

  if l > 0 then
    SetString(Result, buffer, l - 1)
  else
    Result := default;
end;

function GetWindowsLanguageWideName(locale: Integer; useNative: Boolean): UnicodeString;

  function Process(languageLocaleType: Integer): UnicodeString;
  begin
    Result := GetLocaleStr(locale, languageLocaleType, '');
  end;

begin
  if useNative then
    Result := Process(LOCALE_SNATIVELANGNAME)
  else
    Result := Process(LOCALE_SENGLANGUAGE);
end;

function GetWindowsLocaleWideName(locale: Integer; useNative: Boolean): UnicodeString;

  function Process(languageLocaleType, countryLocaleType: Integer): UnicodeString;
  var
    str: UnicodeString;
  begin
    Result := GetLocaleStr(locale, languageLocaleType, '');

    if LocaleToSub(locale) <> SUBLANG_NEUTRAL then
    begin
      str := GetLocaleStr(locale, countryLocaleType, '');

      if str <> '' then
        Result := Result + ' (' + str + ')';
    end;
  end;

begin
  if useNative then
    Result := Process(LOCALE_SNATIVELANGNAME, LOCALE_SNATIVECTRYNAME)
  else
    Result := Process(LOCALE_SENGLANGUAGE, LOCALE_SENGCOUNTRY);
end;

function LocaleEnumProc(localeString: PChar): Integer; stdcall;
var
  locale: Integer;
  str, language, country: String;
begin
  str := localeString;
  locale := StrToInt('$' + str);
  Result := 1;

  language := GetLocaleStr(locale, LOCALE_SISO639LANGNAME, '');
  country := GetLocaleStr(locale, LOCALE_SISO3166CTRYNAME, '');

  if (CompareText(language, FLanguage) = 0) and ((FCountry = '') or (CompareText(country, FCountry) = 0)) then
  begin
    if FCountry = '' then
      FLocale := LocaleToPrimary(locale)
    else
      FLocale := locale;

    Result := 0;
  end;
end;

function Iso639ToLocale(
  const language: String;
  const country: String): Integer;
begin
  FLanguage := language;
  FCountry := country;
  FLocale := 0;

  EnumSystemLocales(@LocaleEnumProc, LCID_SUPPORTED);

  Result := FLocale;
end;

function IsoToLocale(const locale: String): Integer;
var
  language, country, variant: String;
begin
  ParseLocaleString(locale, language, country, variant);
  Result := Iso639ToLocale(language, country);
end;

function ExtensionLocaleEnumProc(localeString: PChar): Integer; stdcall;
var
  locale: Integer;
  str, extension: String;
{$IFDEF VCL13}
  extension2: String;
{$ENDIF}
begin
  str := localeString;
  locale := StrToInt('$' + str);

  extension := GetLocaleStr(locale, LOCALE_SABBREVLANGNAME, '');
{$IFDEF VCL13}
  extension2 := GetLocaleStr(locale, LOCALE_SISO639LANGNAME, '') + '-' + GetLocaleStr(locale, LOCALE_SISO3166CTRYNAME, '');
{$ENDIF}

  if (CompareText(extension, FExtension) = 0) {$IFDEF VCL13}or (CompareText(extension2, FExtension) = 0){$ENDIF} then
  begin
    FLocale := locale;
    Result := 0;
  end
  else
    Result := 1;
end;

function ExtensionLanguageEnumProc(localeString: PChar): Integer; stdcall;
var
  locale: Integer;
  str, extension: String;
{$IFDEF VCL13}
  extension2: String;
{$ENDIF}
begin
  str := localeString;
  locale := StrToInt('$' + str);

  extension := GetLocaleStr(locale, LOCALE_SABBREVLANGNAME, '');
  Delete(extension, Length(extension), 1);

{$IFDEF VCL13}
  extension2 := GetLocaleStr(locale, LOCALE_SISO639LANGNAME, '');
{$ENDIF}

  if (CompareText(extension, FExtension) = 0) {$IFDEF VCL13}or (CompareText(extension2, FExtension) = 0){$ENDIF} then
  begin
    FLocale := LocaleToPrimary(locale);
    Result := 0;
  end
  else
    Result := 1;
end;

function ExtensionToLocale(const value: String): Integer;
begin
  Result := 0;

  if value = '' then
    Exit;

  FExtension := value;

  if FExtension[1] = '.' then
    Delete(FExtension, 1, 1);

  FLocale := 0;
  EnumSystemLocales(@ExtensionLocaleEnumProc, LCID_SUPPORTED);

  if FLocale = 0 then
    EnumSystemLocales(@ExtensionLanguageEnumProc, LCID_SUPPORTED);

  Result := FLocale;
end;

function IsResourceFileLoaded: Boolean;
begin
  Result :=
    (LibModuleList <> nil) and
    (LibModuleList.ResInstance <> 0) and
    (LibModuleList.ResInstance <> LibModuleList.Instance);
end;

function GetActiveResourceLocale(defaultLocale: String): String;
begin
  Result := LoadedResourceLocale;

  if Result = '' then
    Result := defaultLocale;
end;

function GetWindowsLanguage: Integer;
type
  TGetDefaultUILanguage = function: LANGID; stdcall;
var
  module: THandle;
  getDefaultUILanguage: TGetDefaultUILanguage;
begin
  module := LoadLibrary('Kernel32.dll');
  try
    getDefaultUILanguage := GetProcAddress(module, 'GetUserDefaultUILanguage');

    if Assigned(getDefaultUILanguage) then
      Result := getDefaultUILanguage
    else
      Result := 0;
  finally
    FreeLibrary(module);
  end;
end;

var
  languageIds: TList;

function ResourceEnumResourceLanguagesProc(
  instance: THandle;
  resType: PChar;
  resName: PChar;
  languageId: Word;
  param: DWord): Bool; stdcall;
begin
  languageIds.Add(Pointer(languageId));
  Result := True;
end;

function GetResourceLocale(resType, resName: PChar; instance: THandle): Integer;
begin
  if instance = 0 then
   instance := HInstance;

  languageIds := TList.Create;
  try
    EnumResourceLanguages(instance, resType, resName, @ResourceEnumResourceLanguagesProc, 0);

    if languageIds.Count > 0 then
      Result := Integer(languageIds[0])
    else
      Result := 0;
  finally
    languageIds.Free;
  end;
end;

function EnumResourceLanguagesProc(
  instance: THandle;
  resType: PChar;
  resName: PChar;
  languageId: Word;
  param: DWord): Bool; stdcall;
begin
  languageIds.Add(Pointer(languageId));
  Result := True;
end;

function EnumResourceNamesProc(
  instance: THandle;
  resType: PChar;
  resName: PChar;
  param: DWord): Bool; stdcall;
var
  i, id, windowsLocale: Integer;
  reader: TReader;
  stream: TResourceStream;
  header: array[0..3] of AnsiChar;
begin
  Result := True;
  stream := TResourceStream.Create(instance, resName, resType);
  reader := TReader.Create(stream, stream.Size);
  try
    if stream.Size < 4 then
      Exit;

    reader.Read(header, 4);

    if header = VCL_FORM_HEADER_C then
    begin
      languageIds := TList.Create;
      try
        EnumResourceLanguages(HInstance, resType, resName, @EnumResourceLanguagesProc, 0);

        if languageIds.Count > 0 then
        begin
          Result := False;

          if languageIds.Count = 1 then
            FFormResourceLocale := Integer(languageIds[0])
          else
          begin
            windowsLocale := GetWindowsLanguage;

            for i  := 0 to languageIds.Count - 1 do
            begin
              id := Integer(languageIds[0]);

              if windowsLocale = id then
              begin
                FFormResourceLocale := id;
                Exit;
              end;
            end;

            windowsLocale := LocaleToPrimary(GetWindowsLanguage);

            for i  := 0 to languageIds.Count - 1 do
            begin
              id := Integer(languageIds[0]);

              if windowsLocale = id then
              begin
                FFormResourceLocale := id;
                Exit;
              end;
            end;
          end;
        end;
      finally
        languageIds.Free;
      end;
    end;
  finally
    reader.Free;
    stream.Free;
  end;
end;

function GetFormLocale: Integer;
begin
  // If form resource locale is not already found enum form resources and
  // get the language of the first form
  if FFormResourceLocale = 0 then
    EnumResourceNames(HInstance, PChar(RT_RCDATA), @EnumResourceNamesProc, 0);

  Result := FFormResourceLocale
end;

procedure InitializeStaticLocale(default: Integer);
var
  locale: Integer;
begin
  if IsResourceFileLoaded then
    Exit;

  locale := GetFormLocale;

  if locale = 0 then
    locale := default;

  SetThreadLocale(locale);
  UpdateFormatSettings(locale);
end;

function GetActiveLocale: String;
begin
  if IsResourceFileLoaded then
    Result := LoadedResourceLocale
  else
    Result := LocaleToExtension(GetFormLocale);
end;

function IsActiveLocaleBidi: Boolean;
begin
  Result := IsLocaleBiDi(GetActiveLocale);
end;

function GetPreviousLocale: String;
begin
  if PreviouslyLoadedResourceLocale <> '' then
    Result := PreviouslyLoadedResourceLocale
  else
    Result := LocaleToExtension(GetFormLocale);
end;

function IsPreviousLocaleBidi: Boolean;
begin
  Result := IsLocaleBiDi(GetPreviousLocale);
end;

function IsActiveLocaleAsian: Boolean;
begin
  Result := IsLocaleAsian(GetActiveLocale);
end;

procedure InitializeForm(form: TCustomForm);
begin
  if IsActiveLocaleBidi then
    form.FlipChildren(True);
end;

procedure InitializeApplication(layout: TLaUiLayout);
begin
  // Sets the bidi mode of the application
  if IsActiveLocaleBidi then
    Application.BiDiMode := bdRightToLeft
  else
    Application.BiDiMode := bdLeftToRight;

  // Sets the initial layout of the application
  if GetActiveLocale = '' then
    UiLayout := layout
  else if IsActiveLocaleBidi then
    UiLayout := ulRightToLeft
  else
    UiLayout := ulLeftToRight;
end;

function GetResourceFileFileName(
  const fileName: String;
  locale: Integer): String;

  function Process(const fileName, ext: String; locale: Integer): String;
  begin
    Result := ChangeFileExt(fileName, ext);

    if LoadLibraryEx(PChar(Result), 0, LOAD_LIBRARY_AS_DATAFILE) = 0 then
      Result := '';
  end;

var
  ext, langExt: String;
begin
  ext := '.' + GetLocaleStr(locale, LOCALE_SABBREVLANGNAME, '');

  langExt := ext;
  Delete(langExt, Length(langExt), 1);

  if LocaleToSub(locale) = SUBLANG_NEUTRAL then
  begin
    Result := Process(fileName, langExt, locale);

    if Result = '' then
      Result := Process(fileName, ext, locale);
  end
  else
  begin
    Result := Process(fileName, ext, locale);

    if Result = '' then
      Result := Process(fileName, langExt, LocaleToPrimary(locale));
  end;
end;

function LoadNewResourceFile(const code: String): Integer;

  function Process(const fileName, ext: String): Integer;
  var
    dllFileName: String;
  begin
    dllFileName := ChangeFileExt(fileName, ext);
    Result := LoadLibraryEx(PChar(dllFileName), 0, LOAD_LIBRARY_AS_DATAFILE);

    if Result = 0 then
    begin
      dllFileName := GetFolderPath(CSIDL_PERSONAL) + '\' + APPLICATION_DIR + '\' + ChangeFileExt(ExtractFileName(fileName), ext);
      Result := LoadLibraryEx(PChar(dllFileName), 0, LOAD_LIBRARY_AS_DATAFILE);
    end;

    if Result <> 0 then
    begin
      PreviouslyLoadedResourceLocale := LoadedResourceLocale;
      LoadedResourceLocale := ext;
      Delete(LoadedResourceLocale, 1, 1);
    end;
  end;

var
  newInst: Longint;
  fileName, ext: String;
  module: PLibModule;
  buffer: array[0..260] of Char;
begin
  Result := 0;
  module := LibModuleList;

  while module <> nil do
  begin
    if code = '' then
    begin
      if module.ResInstance <> module.Instance then
        FreeLibrary(module.ResInstance);

      module.ResInstance := module.Instance;
      Result := module.Instance;

      PreviouslyLoadedResourceLocale := LoadedResourceLocale;
      LoadedResourceLocale := '';
    end
    else
    begin
      GetModuleFileName(module.Instance, buffer, SizeOf(buffer));
      fileName := buffer;

      if FileExists(fileName) then
      begin
        ext := '.' + code;
        newInst := Process(fileName, ext);

        if (newInst = 0) and (Length(code) = 3) then
        begin
          Delete(ext, Length(ext), 1);
          newInst := Process(fileName, ext);
        end;

        if newInst <> 0 then
        begin
          if module.ResInstance <> module.Instance then
            FreeLibrary(module.ResInstance);

          module.ResInstance := newInst;

          if Result = 0 then
            Result := newInst;
        end;
      end;
    end;

    module := module.Next;
  end;
end;

function GetLanguageFileExtension(locale: Integer): String;
begin
  if locale = 0 then
    Result := ''
  else
  begin
    Result := GetLocaleStr(locale, LOCALE_SABBREVLANGNAME, '');

    if LocaleToSub(locale) = 0 then
      Delete(Result, Length(Result), 1);
  end;
end;

function GetRunningFileName: String;
var
  len: Integer;
begin
  SetLength(Result, MAX_PATH);
  len := GetModuleFileName(HInstance, PChar(Result), MAX_PATH);

  if len > 0 then
    SetLength(Result, len)
  else
    Result := Application.ExeName;
end;

function GetCurrentLanguageFile: String;
begin
  Result := GetLanguageFile(GetRunningFileName, LoadedResourceLocale);
end;

function GetLanguageFile(const exeFileName: String; const code: String): String;
begin
  Result := ChangeFileExt(exeFileName, '.' + code);
end;

function GetAvailableResourceCount(
  exeFileName: String;
  compatibleOnly: Boolean): Integer;
var
  list: TList;
begin
  if exeFileName = '' then
    exeFileName := GetRunningFileName;

  list := TList.Create;
  try
    GetAvailableResources(list, exeFileName, compatibleOnly);
    Result := list.Count;
  finally
    list.Free;
  end;
end;

var
  enumList: TStrings;
  enumExeFileName: String;
  enumCompatibleOnly: Boolean;

function EnumLocales(localeStr: PAnsiChar): Integer; stdcall;

  procedure Process(
    const code: String;
    id: Integer;
    checkIfExists: Boolean);
  var
    i: Integer;
    fileName: String;
  begin
    fileName := ChangeFileExt(enumExeFileName, '.' + code);

    if FileExists(fileName) and (not enumCompatibleOnly or IsLocaleCompatible(id)) then
    begin
      if checkIfExists then
        for i := 0 to enumList.Count - 1 do
          if enumList[i] = code then
            Exit;

      enumList.AddObject(code, TObject(id));
    end;
  end;

var
  id, primary: Integer;
  code: String;
{$IFDEF UNICODE_VCL}
  ansiStr: AnsiString;
{$ENDIF}
{$IFDEF VCL13}
  languageCode: String;
{$ENDIF}
begin
{$IFDEF UNICODE_VCL}
  ansiStr := localeStr;
  id := StrToInt('$' + String(ansiStr));
{$ELSE}
  id := StrToInt('$' + localeStr);
{$ENDIF}

  primary := LocaleToPrimary(id);

{$IFDEF VCL13}
  languageCode := GetLocaleStr(id, LOCALE_SISO639LANGNAME, '');
  Process(languageCode, primary, True);

  code := languageCode + '-' + GetLocaleStr(id, LOCALE_SISO3166CTRYNAME, '');
  Process(code, id, False);
{$ENDIF}

  code := GetLocaleStr(id, LOCALE_SABBREVLANGNAME, '');
  Process(code, id, False);

  Delete(code, Length(code), 1);

{$IFDEF VCL13}
  if not SameText(code, languageCode) then
{$ENDIF}
    Process(code, primary, True);

  Result := 1;
end;

function GetAvailableResources(
  languages: TStrings;
  exeFileName: String;
  compatibleOnly: Boolean): Integer;
begin
  if exeFileName = '' then
    exeFileName := GetRunningFileName;

  enumList := languages;
  enumExeFileName := exeFileName;
  enumCompatibleOnly := compatibleOnly;

  EnumSystemLocalesA(@EnumLocales, LCID_SUPPORTED);

  if languages.Count = 0 then
  begin
    enumExeFileName := GetFolderPath(CSIDL_PERSONAL) + '\' + APPLICATION_DIR + '\' + ExtractFileName(exeFileName);
    EnumSystemLocalesA(@EnumLocales, LCID_SUPPORTED);
  end;

  Result := languages.Count;
end;

function GetAvailableResources(
  languages: TList;
  exeFileName: String;
  compatibleOnly: Boolean): Integer;
var
  i: Integer;
  codes: TStringList;
begin
  codes := TStringList.Create;
  try
    GetAvailableResources(codes, exeFileName, compatibleOnly);

    for i := 0 to codes.Count - 1 do
      languages.Add(codes.Objects[i]);

    Result := languages.Count;
  finally
    codes.Free;
  end;
end;

procedure CheckResourceCount;
begin
  if GetAvailableResourceCount = 0 then
    raise Exception.Create('In order to change language you must first create resource files.');
end;

procedure DisableResourceDllUse;
begin
  SetNewResourceFile('');
end;

function SetNewResourceFile(
  const code: String;
  options: TLaResourceOptions;
  const originalCode: String): Boolean;
var
  locale: Integer;
begin
  LaResourceOptions := options;
  Result := LoadNewResourceFile(code) <> 0;

  if Result then
  begin
    if code = '' then
      locale := ExtensionToLocale(originalCode)
    else
      locale := ExtensionToLocale(code);

    // Updates thred's locale, format settings and bidi mode to match
    if not (roNoThreadLocale in options) then
      SetThreadLocale(locale);

    if not (roNoLocaleVariables in options) then
      UpdateFormatSettings(locale);

    if not (roNoUpdateBidiMode in options) then
    begin
      if IsLocaleBiDi(LocaleToExtension(locale)) then
        Application.BiDiMode := bdRightToLeft
      else
        Application.BiDiMode := bdLeftToRight;
    end;

    // Translate forms
    TLaTranslator.TranslateForms;
  end;

  if roSaveLocale in options then
    SetCurrentDefaultLocaleReg;
end;

function SetNewResourceFile(
  locale: Integer;
  options: TLaResourceOptions;
  originalLocale: Integer): Boolean;
begin
  Result := SetNewResourceFile(LocaleToExtension(locale), options, LocaleToExtension(originalLocale));
end;

procedure SetLocaleParameter(const optionChars: String);
var
  locale: Integer;
begin
  locale := IsoToLocale(GetLocaleParameter);

  if locale <> 0 then
    SetNewResourceFile(LocaleToExtension(locale));
end;

function GetFileDateTime(const fileName: String): TDateTime;
{$IFNDEF VCL10}
var
  time: Integer;
{$ENDIF}
begin
{$IFDEF VCL10}
  FileAge(fileName, Result);
{$ELSE}
  time := FileAge(fileName);

  if time > 0 then
    Result := FileDateToDateTime(time)
  else
    Result := 0.0;
{$ENDIF}
end;

procedure SetFileDateTime(const fileName: String; date: TDateTime);
var
  handle: Integer;
begin
  handle := FileOpen(fileName, fmOpenReadWrite);
  try
    FileSetDate(handle, DateTimeToFileDate(date));
  finally
    FileClose(handle);
  end
end;

function ExtractEnumResourceNamesProc(
  instance: THandle;
  resType: PChar;
  resName: PChar;
  param: DWord): Bool; stdcall;
var
  size: Integer;
  unload: Boolean;
  str, fileName: String;
  pointer: PByte;
  resource: THandle;
  stream: TFileStream;
begin
  str := resName;
  fileName := ChangeFileExt(FFileName, '.' + str);

  if (not FileExists(fileName) or
    not (eoCheckDate in FOptions) or
    (GetFileDateTime(fileName) <> FFileDate)) then
  begin
    unload := resName = LoadedResourceLocale;

    if unload then
      LoadNewResourceFile;

    resource := FindResource(instance, PChar(resName), PChar(resType));
    pointer := LockResource(LoadResource(instance, resource));
    size := SizeofResource(instance, resource);

    DeleteFile(fileName);

    stream := TFileStream.Create(fileName, fmCreate);
    try
      stream.Write(pointer^, size);
    finally
      stream.Free;
    end;

    SetFileDateTime(fileName, FFileDate);

    if unload then
      LoadNewResourceFile(resName);

    if eoRemoveFiles in FOptions then
      FExtractedResourceFiles.Add(fileName);

    Inc(FResourceFileCount);
  end;

  Result := True;
end;

function ExtractResourceFiles(
  options: TExtractOptions;
  resourceName: String): Integer;
var
  dir: String;
begin
  if HasWriteAccess(ExtractFileDir(GetRunningFileName)) then
    FFileName := GetRunningFileName
  else
  begin
    dir := GetFolderPath(CSIDL_PERSONAL) + '\' + APPLICATION_DIR;
    CreateDir(dir);
    FFileName := dir + '\' + ExtractFileName(GetRunningFileName);
  end;

  if resourceName = '' then
    resourceName := APPLICATION_RESOURCE;

  FResourceFileCount := 0;
  FFileDate := GetFileDateTime(GetRunningFileName);
  FOptions := options;

  EnumResourceNames(HInstance, PChar(resourceName), @ExtractEnumResourceNamesProc, 0);

  Result := FResourceFileCount;

  if not IsResourceFileLoaded or (FExtractedResourceFiles.Count > 0) then
    SetNewResourceFile(LocaleToExtension(GetUserDefaultLCID));
end;

procedure SetCurrentDefaultLocaleReg;
var
  fileName: array[0..MAX_PATH] of Char;
begin
  if GetModuleFileName(0, fileName, SizeOf(fileName)) > 0 then
    SetDefaultLocaleReg(fileName, LoadedResourceLocale);
end;

function GetCurrentDefaultLocaleReg: String;
var
  fileName: array[0..MAX_PATH] of Char;
begin
  if GetModuleFileName(0, fileName, SizeOf(fileName)) > 0 then
    Result := GetDefaultLocaleReg(fileName)
  else
    Result := '';
end;

procedure SetDefaultLocaleReg(const fileName: String; const code: String);
var
  reg: TRegistry;
begin
  reg := TRegistry.Create;
  try
    reg.RootKey := HKEY_CURRENT_USER;

    if reg.OpenKey(LOCALE_OVERRIDE_KEY, True) then
      reg.WriteString(fileName, code);
  finally
    reg.Free;
  end;
end;

function GetDefaultLocaleReg(const fileName: String): String;
{$IFNDEF VCL14}
var
  keyExists: Boolean;
{$ENDIF}
begin
{$IFDEF VCL14}
  Result := GetLocaleOverride(fileName);
{$ELSE}
  Result := GetDefaultLocaleReg(fileName, keyExists);
{$ENDIF}
end;

function GetDefaultLocaleReg(const fileName: String; out keyExists: Boolean): String;
var
  reg: TRegistry;

  function Open(rootKey: HKEY; const key: String): Boolean;
  begin
    reg.RootKey := rootKey;
    Result := reg.OpenKey(key, False);
  end;

begin
  Result := '';
  keyExists := False;

  reg := TRegistry.Create;
  try
    if Open(HKEY_CURRENT_USER, LOCALE_OVERRIDE_KEY) or
      Open(HKEY_LOCAL_MACHINE, LOCALE_OVERRIDE_KEY)
{$IFDEF UNICODE_VCL}
      or Open(HKEY_CURRENT_USER, OLD_LOCALE_OVERRIDE_KEY)
{$ENDIF}
    then
    begin
      if reg.ValueExists(fileName) then
      begin
        Result := reg.ReadString(fileName);
        keyExists := True;
      end;
    end;
  finally
    reg.Free;
  end;
end;

function ClearCurrentDefaultLocaleReg: Boolean;
var
  fileName: array[0..MAX_PATH] of Char;
begin
  if GetModuleFileName(0, fileName, SizeOf(fileName)) = 0 then
    Result := False
  else
    Result := ClearDefaultLocaleReg(fileName);
end;

function ClearDefaultLocaleReg(const fileName: String): Boolean;
var
  reg: TRegistry;
begin
  Result := False;

  reg := TRegistry.Create;
  try
    reg.RootKey := HKEY_CURRENT_USER;
    
    if reg.OpenKey(LOCALE_OVERRIDE_KEY, False) then
    begin
      Result := reg.ValueExists(fileName);
      reg.DeleteValue(fileName);
    end;
  finally
    reg.Free;
  end;
end;

procedure CopyResourceToFile(
  module: Integer;
  const resName, resType, fileName: String);
var
  resource: THandle;
  pointer: PByte;
  size: Integer;
  stream: TFileStream;
begin
  resource := FindResource(module, PChar(resName), PChar(resType));
  
  if resource = 0 then
    raise Exception.CreateFmt('%s.%s resource was not found', [resType, resName]);

  pointer := LockResource(LoadResource(module, resource));
  size := SizeofResource(module, resource);

  stream := TFileStream.Create(fileName, fmCreate);
  try
    stream.Write(pointer^, size);
  finally
    stream.Free;
  end;
end;

function IsNtOrLater: Boolean;
var
  versionInfo: TOSVersionInfo;
begin
  versionInfo.dwOSVersionInfoSize := Sizeof(versionInfo);
  GetVersionEx(versionInfo);
  Result := versionInfo.dwPlatformId = 2;
end;

function HasWriteAccess(const dir: String): Boolean;
var
  f: TextFile;
  fileName, parentDir: String;
  buffer: array[0..MAX_PATH] of Char;
begin
  try
    if DirectoryExists(dir) then
    begin
      GetTempFileName(PChar(dir), 'slc', 0, buffer);
      fileName := buffer;

      AssignFile(f, fileName);
      Rewrite(f);
      try
        Write(f, 10);
      finally
        Close(f);
        DeleteFile(fileName);
      end;
    end
    else
    begin
      parentDir := ExtractFileDir(dir);
      try
        ChDir(parentDir);
        CreateDir(dir);
      finally
        RemoveDir(dir);
      end;
    end;

    Result := True;
  except
    Result := False;
  end;
end;

function GetFolderPath(nFolder: Integer): String;
var
  handle: THandle;
  proc: TSHGetFolderPath;
  buffer: array[0..MAX_PATH] of Char;
begin
  handle := LoadLibrary('shfolder.dll');
  try
{$IFDEF UNICODE_VCL}
    proc := GetProcAddress(handle, 'SHGetFolderPathW');
{$ELSE}
    proc := GetProcAddress(handle, 'SHGetFolderPathA');
{$ENDIF}

    if Assigned(proc) and (proc(0, nFolder, 0, 0, buffer) = S_OK) then
      Result := buffer
    else
      Result := '';
  finally
    FreeLibrary(handle);
  end;
end;


type
  TVersionInfo = record
    wLength: Word;
    wValueLength: Word;
    wType: Word;
    szKey: array[0..14] of WideChar;
    padding1: array[0..1] of Word;
    value: TVSFixedFileInfo;
    padding2: array[0..1] of Word;
  end;
  PVersionInfo = ^TVersionInfo;

var
  enumFound: Boolean;
  enumVersion1, enumVersion2, enumVersion3, enumVersion4: Integer;

function EnumEnumVersionProcLanguagesProc(
  module: Integer;
  resType: PChar;
  resName: PChar;
  language: Word;
  param: Integer): Bool; stdcall;
var
  resource: HRSRC;
  versionInfo: PVersionInfo;
begin
  resource := FindResource(module, resName, resType);
  versionInfo := LockResource(LoadResource(module, resource));

  enumVersion1 := versionInfo.value.dwFileVersionMS shr 16;
  enumVersion2 := $FF and versionInfo.value.dwFileVersionMS;
  enumVersion3 := versionInfo.value.dwFileVersionLS shr 16;
  enumVersion4 := $FF and versionInfo.value.dwFileVersionLS;

  enumFound := True;

  Result := Bool(0);
end;

function EnumVersionProc(
  module: Integer;
  resType: PChar;
  resName: PChar;
  param: Integer): Bool; stdcall;
begin
  EnumResourceLanguages(module, resType, resName, @EnumEnumVersionProcLanguagesProc, 0);

  Result := Bool(0);
end;

function DoesResourceFileVersionMatch(resInstance: LongWord): Boolean;
var
  exeVersion1, exeVersion2, exeVersion3, exeVersion4: Integer;
begin
  Result := True;

  enumFound := False;
  EnumResourceNames(HInstance, RT_VERSION, @EnumVersionProc, 0);

  if not enumFound then
    Exit;

  exeVersion1 := enumVersion1;
  exeVersion2 := enumVersion2;
  exeVersion3 := enumVersion3;
  exeVersion4 := enumVersion4;

  enumFound := False;
  EnumResourceNames(resInstance, RT_VERSION, @EnumVersionProc, 0);

  if not enumFound then
    Exit;

  Result :=
    (exeVersion1 = enumVersion1) and
    (exeVersion2 = enumVersion2) and
    (exeVersion3 = enumVersion3) and
    (exeVersion4 = enumVersion4);
end;

function DoesLocaleResourceFileVersionMatch(const code: String): Boolean;
var
  instance: HModule;
begin
  instance := LoadLibraryEx(
    PChar(GetLanguageFile(GetRunningFileName, code)),
    0,
    LOAD_LIBRARY_AS_DATAFILE);
  try
    Result := DoesResourceFileVersionMatch(instance);
  finally
    FreeLibrary(instance);
  end;
end;

function DoesResourceFileVersionMatch: Boolean;
begin
  if IsResourceFileLoaded then
    Result := DoesResourceFileVersionMatch(LibModuleList.ResInstance)
  else
    Result := True;
end;

function GetActiveFontCharset: TFontCharset;
begin
  Result := GetCharSetInfo(ExtensionToLocale(LoadedResourceLocale)).charSet;
end;

function GetActiveCodePage: Integer;
begin
  if LoadedResourceLocale = '' then
    Result := 0
  else
    Result := LocaleToCodePage(ExtensionToLocale(LoadedResourceLocale));
end;

function GetActiveLangId(def: Integer): Integer;
begin
  if LoadedResourceLocale = '' then
    Result := def
  else
    Result := ExtensionToLocale(LoadedResourceLocale);
end;

function GetCharSetInfo(langId: Integer): TLaCharsetInfo;
const
  CHARSETSETS: array[0..31] of TLaCharsetInfo =
  (
    (charSet: ANSI_CHARSET; codePage: 1252),
    (charSet: EASTEUROPE_CHARSET; codePage: 1250),
    (charSet: RUSSIAN_CHARSET; codePage: 1251),
    (charSet: GREEK_CHARSET; codePage: 1253),
    (charSet: TURKISH_CHARSET; codePage: 1254),
    (charSet: HEBREW_CHARSET; codePage: 1255),
    (charSet: ARABIC_CHARSET; codePage: 1256),
    (charSet: BALTIC_CHARSET; codePage: 1257),
    (charSet: 0; codePage: 0),
    (charSet: 0; codePage: 0),
    (charSet: 0; codePage: 0),
    (charSet: 0; codePage: 0),
    (charSet: 0; codePage: 0),
    (charSet: 0; codePage: 0),
    (charSet: 0; codePage: 0),
    (charSet: 0; codePage: 0),
    (charSet: THAI_CHARSET; codePage: 874),
    (charSet: SHIFTJIS_CHARSET; codePage: 932),
    (charSet: GB2312_CHARSET; codePage: 936),
    (charSet: HANGEUL_CHARSET; codePage: 949),
    (charSet: CHINESEBIG5_CHARSET; codePage: 950),
    (charSet: JOHAB_CHARSET; codePage: 1361),
    (charSet: VIETNAMESE_CHARSET; codePage: 1258),
    (charSet: 0; codePage: 0),
    (charSet: 0; codePage: 0),
    (charSet: 0; codePage: 0),
    (charSet: 0; codePage: 0),
    (charSet: 0; codePage: 0),
    (charSet: 0; codePage: 0),
    (charSet: 0; codePage: 0),
    (charSet: 0; codePage: 0),
    (charSet: 0; codePage: 0)
  );
var
  f: DWord;
  i: Integer;
  lfs: TLocaleFontSignature;
begin
  if GetLocaleInfo(langId, LOCALE_FONTSIGNATURE, PChar(@lfs), Sizeof(lfs)) <> 0 then
  begin
    f := 1;

    for i := Low(CHARSETSETS) to High(CHARSETSETS) do
    begin
      if (f and lfs.fsCsbDefault[0]) <> 0 then
      begin
        Result := CHARSETSETS[i];
        Exit;
      end;

      f := f shl 1;
    end;
  end;

  Result := CHARSETSETS[0];
end;

constructor TLanguageListItem.Create(const code: String; const name: UnicodeString);
begin
  inherited Create;
  FCode := code;
  FName := name;
end;

function TLanguageListItem.GetCharset: TFontCharset;
begin
  Result := GetCharSetInfo(FLocale).charSet;
end;

class procedure TLanguageListItem.DrawItem(
  parent: TWinControl;
  listBox: TListBox;
  index: Integer;
  rect: TRect;
  state: TOwnerDrawState);

  function TextExtent(const text: UnicodeString): TSize;
  begin
    Result.cX := 0;
    Result.cY := 0;

    GetTextExtentPoint32W(
      listBox.Canvas.Handle,
      PWideChar(text),
      Length(text),
      Result);
  end;

var
  language: TLanguageListItem;
  str: UnicodeString;
  x, options: Integer;
begin
  if IsNtOrLater then
  begin
    x := rect.Left + 2;
    language := TLanguageListItem(listBox.Items.Objects[index]);
    str := language.Name;

    options := ETO_CLIPPED or listBox.Canvas.TextFlags;

    if parent.Brush.Style <> bsClear then
      options := options or ETO_OPAQUE;

    if ((listBox.Canvas.TextFlags and ETO_RTLREADING) <> 0) and
      (listBox.Canvas.CanvasOrientation = coRightToLeft) then
    begin
      Inc(x, TextExtent(str).cx + 1);
    end;

    listBox.Canvas.Font.Charset := language.Charset;

    ExtTextOutW(
      listBox.Canvas.Handle,
      x,
      rect.Top,
      options,
      @rect,
      PWideChar(str),
      Length(str),
      nil);
  end
  else
    listBox.Canvas.TextRect(
      rect,
      rect.Left + 2,
      rect.Top,
      listBox.Items[index]);
end;

procedure DeleteExtractedResourceFiles;
var
  libModule: PLibModule;
begin
  // Do not place this code inside the finalization block
  if (FExtractedResourceFiles <> nil) and (FExtractedResourceFiles.Count > 0) then
  begin
    libModule := LibModuleList;

    while libModule <> nil do
    begin
      if libModule.ResInstance <> libModule.Instance then
        FreeLibrary(libModule.ResInstance);

      libModule.ResInstance := libModule.Instance;
      libModule := libModule.Next;
    end;

    while FExtractedResourceFiles.Count > 0 do
    begin
      DeleteFile(FExtractedResourceFiles[0]);
      FExtractedResourceFiles.Delete(0);
    end;
  end;
end;

const
  KERNEL_C = 'kernel32.dll';

function GetUserDefaultUILanguage: WORD; stdcall; external KERNEL_C name 'GetUserDefaultUILanguage';
function GetSystemDefaultUILanguage: WORD; stdcall; external KERNEL_C name 'GetSystemDefaultUILanguage';

procedure SetInitialResourceLocale;

  function CheckLocale(fileName: String; locale: String; tryPrimary: Boolean): String;
  var
    ext: String;
  begin
    ext := '.' + locale;

    if FileExists(ChangeFileExt(fileName, ext)) then
      Result := locale
    else if (Length(locale) > 2) or tryPrimary then
    begin
      Delete(ext, Length(ext), 1);

      if FileExists(ChangeFileExt(fileName, ext)) then
      begin
        Result := locale;
        Delete(Result, Length(Result), 1);
      end
      else
        Result := '';
    end
    else
      Result := '';
  end;

{$IFDEF VCL13}
  function CheckLocales(fileName: String; locales: String): String;
  var
    p, lang: PChar;
  begin
    p := PChar(locales);

    while p^ <> #0 do
    begin
      lang := p;

      while (p^ <> ',') and (p^ <> #0) do
        Inc(p);

      if p^ = ',' then
      begin
        p^ := #0;
        Inc(p);
      end;

      Result := CheckLocale(fileName, lang, False);

      if Result <> '' then
      begin
        Result := UpperCase(Result);
        Exit;
      end;
    end;
  end;
{$ENDIF}

var
  buffer: array[0..260] of Char;
begin
  PreviouslyLoadedResourceLocale := '';

  if (LoadedResourceLocale = '') and
    (GetModuleFileName(LibModuleList.Instance, buffer, SizeOf(buffer)) > 0) then
  begin
    LoadedResourceLocale := CheckLocale(buffer, GetCurrentDefaultLocaleReg, False);

    if LoadedResourceLocale = '' then
    begin
{$IFDEF VCL13}
      LoadedResourceLocale := CheckLocales(buffer, GetUILanguages(GetUserDefaultUILanguage));

      if (LoadedResourceLocale = '') and ((GetVersion and $000000FF) < 6) then
        LoadedResourceLocale := CheckLocales(buffer, GetUILanguages(GetSystemDefaultUILanguage));

      if LoadedResourceLocale = '' then
        LoadedResourceLocale := CheckLocale(buffer, LocaleToExtension(GetUserDefaultUILanguage), True);
{$ELSE}
      LoadedResourceLocale := CheckLocale(buffer, LocaleToExtension(GetThreadLocale), True);
{$ENDIF}
    end;
  end;
end;

function GetResourceInstance: LongWord;
begin
  if LibModuleList <> nil then
    Result := LibModuleList.ResInstance
  else
    Result := HInstance;
end;

{$IFDEF VCL6}
  {$IFDEF WIN32}
procedure RemapProcedure(oldProc, newProc: Pointer; var data: TProcedureData);
var
  i: Integer;
  p: Pointer;
  protect1, protect2: DWord;
  address: PAnsiChar;
begin
  if data.Address <> nil then
    Exit;

  if not VirtualProtect(oldProc, PROCEDURE_DATA_SIZE_C, PAGE_EXECUTE_READWRITE, @protect1) then
    RaiseLastOSError;

  p := oldProc;

  if Word(p^) = $25FF then
  begin
    Inc(Integer(p), 2);
    p := Pointer(Pointer(p^)^);

    if not VirtualProtect(oldProc, PROCEDURE_DATA_SIZE_C, protect1, @protect2) then
      RaiseLastOSError;

    oldProc := p;

    if not VirtualProtect(oldProc, PROCEDURE_DATA_SIZE_C, PAGE_EXECUTE_READWRITE, @protect1) then
      RaiseLastOSError;
  end;

  Move(oldProc^, data.Data, PROCEDURE_DATA_SIZE_C);
  data.Address := oldProc;

  i := Integer(newProc) - Integer(p) - PROCEDURE_DATA_SIZE_C + 1;

  address := PAnsiChar(oldProc);
  address[0] := AnsiChar($E9);
  address[1] := AnsiChar(i and 255);
  address[2] := AnsiChar((i shr 8) and 255);
  address[3] := AnsiChar((i shr 16) and 255);
  address[4] := AnsiChar((i shr 24) and 255);

  if not VirtualProtect(address, PROCEDURE_DATA_SIZE_C, protect1, @protect2) then
    RaiseLastOSError;
end;

procedure RestoreProcedure(originalProc: Pointer; data: TProcedureData);
var
  protect1, protect2: Cardinal;
begin
  if data.Address = nil then
    Exit;

  if not VirtualProtect(data.Address, PROCEDURE_DATA_SIZE_C, PAGE_EXECUTE_READWRITE, @protect1) then
    RaiseLastOSError;

  Move(data.Data, data.Address^, PROCEDURE_DATA_SIZE_C);

  if not VirtualProtect(data.Address, PROCEDURE_DATA_SIZE_C, protect1, @protect2) then
    RaiseLastOSError;
end;
  {$ENDIF}
{$ENDIF}

initialization
  UiLayout := ulLeftToRight;
  FFormResourceLocale := 0;

  SetInitialResourceLocale;
  FExtractedResourceFiles := TStringList.Create;
finalization
  DeleteExtractedResourceFiles;
  FExtractedResourceFiles.Free;
end.
